@@ -14,22 +14,23 @@ function ZipFile() {
   this.outputStream = new PassThrough();
   this.entries = [];
   this.outputStreamCursor = 0;
-  this.ended = false;
+  this.ended = false; // .end() sets this
+  this.allDone = false; // set when we've written the last bytes
 }
 
 ZipFile.prototype.addFile = function(realPath, metadataPath, options) {
   var self = this;
-  validateMetadataPath(metadataPath);
+  metadataPath = validateMetadataPath(metadataPath, false);
   if (options == null) options = {};
 
-  var entry = new Entry(metadataPath, options);
+  var entry = new Entry(metadataPath, false, options);
   self.entries.push(entry);
   fs.stat(realPath, function(err, stats) {
     if (err) return self.emit("error", err);
     if (!stats.isFile()) return self.emit("error", new Error("not a file: " + realPath));
     entry.uncompressedSize = stats.size;
-    if (entry.lastModFileTime == null || entry.lastModFileDate == null) entry.setLastModDate(stats.mtime);
-    if (entry.externalFileAttributes == null) entry.setFileAttributesMode(stats.mode);
+    if (options.mtime == null) entry.setLastModDate(stats.mtime);
+    if (options.mode == null) entry.setFileAttributesMode(stats.mode);
     entry.setFileDataPumpFunction(function() {
       var readStream = fs.createReadStream(realPath);
       entry.state = Entry.FILE_DATA_IN_PROGRESS;
@@ -44,27 +45,26 @@ ZipFile.prototype.addFile = function(realPath, metadataPath, options) {
 
 ZipFile.prototype.addReadStream = function(readStream, metadataPath, options) {
   var self = this;
-  validateMetadataPath(metadataPath);
+  metadataPath = validateMetadataPath(metadataPath, false);
   if (options == null) options = {};
-  var entry = new Entry(metadataPath, options);
-  validateFilelessEntryProperties(entry);
+  var entry = new Entry(metadataPath, false, options);
   self.entries.push(entry);
   entry.setFileDataPumpFunction(function() {
     entry.state = Entry.FILE_DATA_IN_PROGRESS;
     pumpFileDataReadStream(self, entry, readStream);
-    pumpEntries(self);
   });
   pumpEntries(self);
 };
 
 ZipFile.prototype.addBuffer = function(buffer, metadataPath, options) {
   var self = this;
-  validateMetadataPath(metadataPath);
+  metadataPath = validateMetadataPath(metadataPath, false);
   if (options == null) options = {};
-  var entry = new Entry(metadataPath, options);
-  validateFilelessEntryProperties(entry, buffer.length);
+  if (options.size != null) throw new Error("options.size not allowed");
+  var entry = new Entry(metadataPath, false, options);
   entry.uncompressedSize = buffer.length;
   entry.crc32 = crc32.unsigned(buffer);
+  entry.crcAndFileSizeKnown = true;
   self.entries.push(entry);
   if (!entry.compress) {
     setCompressedBuffer(buffer);
@@ -79,12 +79,33 @@ ZipFile.prototype.addBuffer = function(buffer, metadataPath, options) {
       writeToOutputStream(self, compressedBuffer);
       writeToOutputStream(self, entry.getFileDescriptor());
       entry.state = Entry.FILE_DATA_DONE;
-      pumpEntries(self);
+
+      // don't call pumpEntries() recursively.
+      // (also, don't call process.nextTick recursively.)
+      setImmediate(function() {
+        pumpEntries(self);
+      });
     });
     pumpEntries(self);
   }
 };
 
+ZipFile.prototype.addEmptyDirectory = function(metadataPath, options) {
+  var self = this;
+  metadataPath = validateMetadataPath(metadataPath, true);
+  if (options == null) options = {};
+  if (options.size != null) throw new Error("options.size not allowed");
+  if (options.compress != null) throw new Error("options.compress not allowed");
+  var entry = new Entry(metadataPath, true, options);
+  self.entries.push(entry);
+  entry.setFileDataPumpFunction(function() {
+    writeToOutputStream(self, entry.getFileDescriptor());
+    entry.state = Entry.FILE_DATA_DONE;
+    pumpEntries(self);
+  });
+  pumpEntries(self);
+};
+
 ZipFile.prototype.end = function(finalSizeCallback) {
   if (this.ended) return;
   this.ended = true;
@@ -123,6 +144,7 @@ function pumpFileDataReadStream(self, entry, readStream) {
 }
 
 function pumpEntries(self) {
+  if (self.allDone) return;
   // first check if finalSize is finally known
   if (self.ended && self.finalSizeCallback != null) {
     var finalSize = calculateFinalSize(self);
@@ -162,6 +184,7 @@ function pumpEntries(self) {
       });
       writeToOutputStream(self, getEndOfCentralDirectoryRecord(self));
       self.outputStream.end();
+      self.allDone = true;
     }
   }
 }
@@ -181,8 +204,8 @@ function calculateFinalSize(self) {
     }
     result += LOCAL_FILE_HEADER_FIXED_SIZE + entry.utf8FileName.length +
               entry.uncompressedSize +
-              FILE_DESCRIPTOR_SIZE +
               CENTRAL_DIRECTORY_RECORD_FIXED_SIZE + entry.utf8FileName.length;
+    if (!entry.crcAndFileSizeKnown) result += FILE_DESCRIPTOR_SIZE;
   }
   result += END_OF_CENTRAL_DIRECTORY_RECORD_SIZE;
   return result;
@@ -203,28 +226,52 @@ function getEndOfCentralDirectoryRecord(self) {
   return buffer;
 }
 
-function validateMetadataPath(metadataPath) {
+function validateMetadataPath(metadataPath, isDirectory) {
+  if (metadataPath === "") throw new Error("empty metadataPath");
   if (metadataPath.indexOf("\\") !== -1) throw new Error("invalid characters in path: " + metadataPath);
   if (/^[a-zA-Z]:/.test(metadataPath) || /^\//.test(metadataPath)) throw new Error("absolute path: " + metadataPath);
   if (metadataPath.split("/").indexOf("..") !== -1) throw new Error("invalid relative path: " + metadataPath);
-}
-function validateFilelessEntryProperties(entry, length) {
-  if (entry.lastModFileTime == null || entry.lastModFileDate == null) throw new Error("missing options.mtime");
-  if (entry.externalFileAttributes == null) throw new Error("missing options.mode");
-  if (entry.uncompressedSize != null && length != null && entry.uncompressedSize !== length) throw new Error("invalid options.size");
+  var looksLikeDirectory = /\/$/.test(metadataPath);
+  if (isDirectory) {
+    // append a trailing '/' if necessary.
+    if (!looksLikeDirectory) metadataPath += "/";
+  } else {
+    if (looksLikeDirectory) throw new Error("file path cannot end with '/': " + metadataPath);
+  }
+  return metadataPath;
 }
 
 // this class is not part of the public API
-function Entry(metadataPath, options) {
+function Entry(metadataPath, isDirectory, options) {
   this.utf8FileName = new Buffer(metadataPath);
   if (this.utf8FileName.length > 0xffff) throw new Error("utf8 file name too long. " + utf8FileName.length + " > " + 0xffff);
+  this.isDirectory = isDirectory;
   this.state = Entry.WAITING_FOR_METADATA;
-  if (options.mtime != null) this.setLastModDate(options.mtime);
-  if (options.mode != null) this.setFileAttributesMode(options.mode);
-  this.uncompressedSize = null; // unknown
-  if (options.size != null) this.uncompressedSize = options.size;
-  this.compress = true; // default
-  if (options.compress != null) this.compress = !!options.compress;
+  this.setLastModDate(options.mtime != null ? options.mtime : new Date());
+  if (options.mode != null) {
+    this.setFileAttributesMode(options.mode);
+  } else {
+    this.setFileAttributesMode(isDirectory ? 040775 : 0100664);
+  }
+  if (isDirectory) {
+    this.crcAndFileSizeKnown = true;
+    this.crc32 = 0;
+    this.uncompressedSize = 0;
+    this.compressedSize = 0;
+  } else {
+    // unknown so far
+    this.crcAndFileSizeKnown = false;
+    this.crc32 = null;
+    this.uncompressedSize = null;
+    this.compressedSize = null;
+    if (options.size != null) this.uncompressedSize = options.size;
+  }
+  if (isDirectory) {
+    this.compress = false;
+  } else {
+    this.compress = true; // default
+    if (options.compress != null) this.compress = !!options.compress;
+  }
 }
 Entry.WAITING_FOR_METADATA = 0;
 Entry.READY_TO_PUMP_FILE_DATA = 1;
@@ -240,6 +287,7 @@ Entry.prototype.setFileAttributesMode = function(mode) {
   // http://unix.stackexchange.com/questions/14705/the-zip-formats-external-file-attribute/14727#14727
   this.externalFileAttributes = (mode << 16) >>> 0;
 };
+// doFileDataPump() should not call pumpEntries() directly. see issue #9.
 Entry.prototype.setFileDataPumpFunction = function(doFileDataPump) {
   this.doFileDataPump = doFileDataPump;
   this.state = Entry.READY_TO_PUMP_FILE_DATA;
@@ -252,17 +300,28 @@ var VERSION_MADE_BY_INFO_ZIP = 0x031e;
 var FILE_NAME_IS_UTF8 = 1 << 11;
 var UNKNOWN_CRC32_AND_FILE_SIZES = 1 << 3;
 Entry.prototype.getLocalFileHeader = function() {
+  var crc32 = 0;
+  var compressedSize = 0;
+  var uncompressedSize = 0;
+  if (this.crcAndFileSizeKnown) {
+    crc32 = this.crc32;
+    compressedSize = this.compressedSize;
+    uncompressedSize = this.uncompressedSize;
+  }
+
   var fixedSizeStuff = new Buffer(LOCAL_FILE_HEADER_FIXED_SIZE);
-  var generalPurposeBitFlag = UNKNOWN_CRC32_AND_FILE_SIZES | FILE_NAME_IS_UTF8;
+  var generalPurposeBitFlag = FILE_NAME_IS_UTF8;
+  if (!this.crcAndFileSizeKnown) generalPurposeBitFlag |= UNKNOWN_CRC32_AND_FILE_SIZES;
+
   fixedSizeStuff.writeUInt32LE(0x04034b50, 0);                  // local file header signature     4 bytes  (0x04034b50)
   fixedSizeStuff.writeUInt16LE(VERSION_NEEDED_TO_EXTRACT, 4);   // version needed to extract       2 bytes
   fixedSizeStuff.writeUInt16LE(generalPurposeBitFlag, 6);       // general purpose bit flag        2 bytes
   fixedSizeStuff.writeUInt16LE(this.getCompressionMethod(), 8); // compression method              2 bytes
   fixedSizeStuff.writeUInt16LE(this.lastModFileTime, 10);       // last mod file time              2 bytes
   fixedSizeStuff.writeUInt16LE(this.lastModFileDate, 12);       // last mod file date              2 bytes
-  fixedSizeStuff.writeUInt32LE(0, 14);                          // crc-32                          4 bytes
-  fixedSizeStuff.writeUInt32LE(0, 18);                          // compressed size                 4 bytes
-  fixedSizeStuff.writeUInt32LE(0, 22);                          // uncompressed size               4 bytes
+  fixedSizeStuff.writeUInt32LE(crc32, 14);                      // crc-32                          4 bytes
+  fixedSizeStuff.writeUInt32LE(compressedSize, 18);             // compressed size                 4 bytes
+  fixedSizeStuff.writeUInt32LE(uncompressedSize, 22);           // uncompressed size               4 bytes
   fixedSizeStuff.writeUInt16LE(this.utf8FileName.length, 26);   // file name length                2 bytes
   fixedSizeStuff.writeUInt16LE(0, 28);                          // extra field length              2 bytes
   return Buffer.concat([
@@ -273,6 +332,10 @@ Entry.prototype.getLocalFileHeader = function() {
 };
 var FILE_DESCRIPTOR_SIZE = 16
 Entry.prototype.getFileDescriptor = function() {
+  if (this.crcAndFileSizeKnown) {
+    // MAC's Archive Utility requires this not be present unless we set general purpose bit 3
+    return new Buffer(0);
+  }
   var buffer = new Buffer(FILE_DESCRIPTOR_SIZE);
   buffer.writeUInt32LE(0x08074b50, 0);             // optional signature (required according to Archive Utility)
   buffer.writeUInt32LE(this.crc32, 4);             // crc-32                          4 bytes
@@ -283,10 +346,13 @@ Entry.prototype.getFileDescriptor = function() {
 var CENTRAL_DIRECTORY_RECORD_FIXED_SIZE = 46;
 Entry.prototype.getCentralDirectoryRecord = function() {
   var fixedSizeStuff = new Buffer(CENTRAL_DIRECTORY_RECORD_FIXED_SIZE);
+  var generalPurposeBitFlag = FILE_NAME_IS_UTF8;
+  if (!this.crcAndFileSizeKnown) generalPurposeBitFlag |= UNKNOWN_CRC32_AND_FILE_SIZES;
+  
   fixedSizeStuff.writeUInt32LE(0x02014b50, 0);                        // central file header signature   4 bytes  (0x02014b50)
   fixedSizeStuff.writeUInt16LE(VERSION_MADE_BY_INFO_ZIP, 4);          // version made by                 2 bytes
   fixedSizeStuff.writeUInt16LE(VERSION_NEEDED_TO_EXTRACT, 6);         // version needed to extract       2 bytes
-  fixedSizeStuff.writeUInt16LE(FILE_NAME_IS_UTF8, 8);                 // general purpose bit flag        2 bytes
+  fixedSizeStuff.writeUInt16LE(generalPurposeBitFlag, 8);             // general purpose bit flag        2 bytes
   fixedSizeStuff.writeUInt16LE(this.getCompressionMethod(), 10);      // compression method              2 bytes
   fixedSizeStuff.writeUInt16LE(this.lastModFileTime, 12);             // last mod file time              2 bytes
   fixedSizeStuff.writeUInt16LE(this.lastModFileDate, 14);             // last mod file date              2 bytes
